import datetime
import typing as t
from ._types import (
    WerkitErrorOrigin,
    WerkitErrorOutputMessage,
    WerkitSuccessOutputMessage,
)


ResultType = t.TypeVar("ResultType")
MessageKeyType = t.TypeVar("MessageKeyType")


def serialize_result(
    message_key: t.Any,
    serializable_result: t.Any,
    start_time: datetime.datetime,
    duration_seconds: float,
    runtime_info: t.Any = None,
) -> WerkitSuccessOutputMessage[ResultType, MessageKeyType]:
    """
    Wrap the computation result in the `werkit.Manager` result schema.

    Args:
        serializable_result (object): Any serializable result.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The computation time.
        runtime_info (object): Serializable runtime metadata.
    """
    return {
        "message_key": message_key,
        "success": True,
        "result": serializable_result,
        "error": None,
        "error_origin": None,
        "start_time": start_time.astimezone().isoformat(),
        "duration_seconds": duration_seconds,
        "runtime_info": runtime_info,
    }


def serialize_exception(
    message_key: t.Any,
    exception: BaseException,
    error_origin: WerkitErrorOrigin,
    start_time: datetime.datetime,
    duration_seconds: float = -1,
    runtime_info: t.Any = None,
) -> WerkitErrorOutputMessage[MessageKeyType]:
    """
    Wrap an exception in the `werkit.Manager` result schema.

    Args:
        exception (Exception): The exception to wrap.
        error_origin (str): Either `"compute"`, `"system"` or
            `"orchestration"`. Refer to the `werkit.Manager` documentation for
            the definitions.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The time elapsed before the error occurred,
            if known.
        runtime_info (object): Serializable runtime metadata.
    """
    import traceback
    from ._synthetic_error import SyntheticError

    if isinstance(exception, SyntheticError):
        error = exception.error
        out_error_origin = exception.error_origin
    else:
        error = traceback.format_exception(None, exception, exception.__traceback__)
        out_error_origin = error_origin

    return {
        "message_key": message_key,
        "success": False,
        "result": None,
        "error": error,
        "error_origin": out_error_origin,
        "duration_seconds": duration_seconds,
        "start_time": start_time.astimezone().isoformat(),
        "runtime_info": runtime_info,
    }



import os
import shutil
import sys
import venv
import zipfile
from pathlib import Path

from executor import execute


def export_poetry_requirements(
    output_file: str,
    extras: list[str] = [],
    with_credentials: bool = True,
    with_hashes: bool = True,
) -> None:
    args = ["poetry", "export", "--output", output_file]
    if with_credentials:
        args.append("--with-credentials")
    if not with_hashes:
        args.append("--without-hashes")
    for extra in extras:
        args += ["--extras", extra]
    execute(*args)


def create_venv_with_dependencies(
    venv_dir: str,
    upgrade_pip: bool = True,
    install_wheel: bool = True,
    install_requirements_from: list[str] = ["requirements.txt"],
    install_transitive_dependencies: bool = True,
    environment: dict[str, str] = {},
) -> None:
    venv.create(venv_dir, with_pip=True)
    python = os.path.join(venv_dir, "bin", "python")

    if upgrade_pip:
        execute(
            python, "-m", "pip", "install", "--upgrade", "pip", environment=environment
        )

    if install_wheel:
        execute(python, "-m", "pip", "install", "wheel", environment=environment)

    if len(install_requirements_from) > 0:
        args = [python, "-m", "pip", "install"]
        if not install_transitive_dependencies:
            args += ["--no-dependencies"]
        for requirements_file in install_requirements_from:
            args += ["-r", requirements_file]
        execute(*args, environment=environment)


def site_packages_for_venv(venv_dir: str) -> str:
    python = os.path.join(venv_dir, "bin", "python")
    return execute(
        python,
        "-c",
        'import sysconfig; print(sysconfig.get_paths()["purelib"])',
        capture=True,
    )


def collect_zipfile_contents(
    target_dir: str,
    venv_dir: str,
    src_files: list[str] = [],
    src_dirs: list[str] = [],
    lib_files: list[str] = [],
    verbose: bool = False,
) -> None:
    def pif(x: str) -> None:
        if verbose:
            print(x, file=sys.stderr)

    if os.path.isdir(target_dir):
        raise ValueError(f"target_dir should not exist: {target_dir}")
    if not os.path.isdir(venv_dir):
        raise ValueError(f"venv_dir should already be populated: {venv_dir}")

    # Copy dependencies from venv.
    site_packages_dir = site_packages_for_venv(venv_dir)
    pif(f"Copying dependencies from {site_packages_dir} to {target_dir}")
    shutil.copytree(site_packages_dir, target_dir)

    if lib_files:
        lib_dir = os.path.join(target_dir, "lib")
        os.makedirs(lib_dir, exist_ok=True)
        for lib_file in lib_files:
            target = os.path.join(lib_dir, os.path.basename(lib_file))
            pif(f"Copying {lib_file} to {target}/")
            shutil.copyfile(lib_file, target)

    for src_dir in src_dirs:
        target = os.path.join(target_dir, src_dir)
        pif(f"Copying {src_dir} to {target}")
        shutil.copytree(src_dir, target)

    for src_file in src_files:
        target = os.path.join(target_dir, src_file)
        Path(os.path.dirname(target)).mkdir(parents=True, exist_ok=True)

        pif(f"Copying {src_file} to {target}")
        shutil.copyfile(src_file, target)


def create_zipfile_from_dir(dir_path: str, path_to_zipfile: str) -> None:
    with zipfile.ZipFile(path_to_zipfile, "w", zipfile.ZIP_DEFLATED) as f:
        for root, dirs, files in os.walk(dir_path):
            for file in files:
                fullpath = os.path.join(root, file)
                arcname = os.path.relpath(fullpath, start=dir_path)
                f.write(fullpath, arcname=arcname)



import os
import sys
import typing as t
import uuid
from contextlib import AbstractContextManager, contextmanager


@contextmanager
def temp_file_on_s3(
    local_path: str, bucket: str, key: t.Optional[str] = None, verbose: bool = False
) -> t.Iterator[str]:
    """
    Copy the given path to S3. Delete the object from S3 when the block
    exits.
    """
    import boto3

    def pif(x: str) -> None:
        if verbose:
            print(x, file=sys.stderr)

    if key is None:
        filename, extension = os.path.splitext(os.path.basename(local_path))
        key = f"{filename}_{uuid.uuid4().hex}{extension}"

    s3_client = boto3.client("s3")

    file_on_s3 = f"s3://{bucket}/{key}"
    pif(f"Uploading {local_path} to {file_on_s3}")
    s3_client.upload_file(Filename=local_path, Bucket=bucket, Key=key)

    try:
        yield key
    finally:
        pif(f"Removing {file_on_s3}")
        s3_client.delete_object(Bucket=bucket, Key=key)


@t.overload
def temp_file_on_s3_from_string(  # noqa: E704
    contents: str,
    bucket: str,
    ret_etag: t.Literal[True],
    key: t.Optional[str] = None,
    extension: t.Optional[str] = None,
    verbose: bool = False,
) -> AbstractContextManager[tuple[str, str]]: ...


@t.overload
def temp_file_on_s3_from_string(  # noqa: E704
    contents: str,
    bucket: str,
    ret_etag: t.Literal[False],
    key: t.Optional[str] = None,
    extension: t.Optional[str] = None,
    verbose: bool = False,
) -> AbstractContextManager[str]: ...


@t.overload
def temp_file_on_s3_from_string(  # noqa: E704
    contents: str,
    bucket: str,
    *,
    ret_etag: bool = False,
    key: t.Optional[str] = None,
    extension: t.Optional[str] = None,
    verbose: bool = False,
) -> AbstractContextManager[str]: ...


# @t.overload doesn't work properly with @contextmanager
# https://github.com/python/mypy/issues/14652
@contextmanager  # type: ignore[arg-type]
def temp_file_on_s3_from_string(  # type: ignore[misc]
    contents: str,
    bucket: str,
    ret_etag: bool = False,
    key: t.Optional[str] = None,
    extension: t.Optional[str] = None,
    verbose: bool = False,
) -> t.Union[AbstractContextManager[str], AbstractContextManager[tuple[str, str]]]:
    """
    Write the given contents to S3. Delete the object from S3 when the block
    exits.
    """
    import boto3

    def pif(x: str) -> None:
        if verbose:
            print(x, file=sys.stderr)

    if key is None:
        key = f"{uuid.uuid4().hex}{extension}"

    s3_client = boto3.client("s3")

    file_on_s3 = f"s3://{bucket}/{key}"
    pif(f"Writing to {file_on_s3}")

    response = s3_client.put_object(
        Bucket=bucket, Key=key, Body=contents.encode("utf-8")
    )

    try:
        yield (key, response["ETag"]) if ret_etag else key
    finally:
        pif(f"Removing {file_on_s3}")
        s3_client.delete_object(Bucket=bucket, Key=key)



from ._types import WerkitErrorOrigin, WerkitErrorOutputMessage


class SyntheticError(Exception):
    """An exception that wraps a serialized error output message.

    This is useful for passing along an error output message that was
    generated elsewhere, such as in a subprocess.

    Only the message's `error` and `error_origin` keys are preserved.
    """

    error: list[str]
    error_origin: WerkitErrorOrigin

    def __init__(self, output_message: WerkitErrorOutputMessage):
        assert output_message["success"] is False
        self.error = output_message["error"]
        self.error_origin = output_message["error_origin"]
        super().__init__("".join(self.error))



import typing as t
from typing_extensions import TypedDict

MessageKeyType = t.TypeVar("MessageKeyType")
ResultType = t.TypeVar("ResultType")


class WerkitComputeMeta(TypedDict):
    start_time: str
    duration_seconds: float
    runtime_info: t.Optional[t.Any]


class WerkitSuccessOutputMessage(
    WerkitComputeMeta, TypedDict, t.Generic[ResultType, MessageKeyType]
):
    success: t.Literal[True]
    result: ResultType
    error: None
    error_origin: None
    message_key: MessageKeyType


WerkitErrorOrigin = t.Literal["compute", "system", "orchestration"]


class WerkitErrorOutputMessage(WerkitComputeMeta, TypedDict, t.Generic[MessageKeyType]):
    success: t.Literal[False]
    result: None
    error: list[str]
    error_origin: WerkitErrorOrigin
    message_key: MessageKeyType


WerkitOutputMessage = t.Union[
    WerkitSuccessOutputMessage[ResultType, MessageKeyType],
    WerkitErrorOutputMessage[MessageKeyType],
]
